@@ -1,18 +1,23 @@
 # https://hadoop.apache.org/docs/r1.0.4/webhdfs.html
 
-import requests
-from urllib.parse import quote
-import uuid
-from ..spec import AbstractFileSystem, AbstractBufferedFile
-from ..utils import infer_storage_options
 import logging
+import secrets
+import shutil
+import uuid
+from contextlib import suppress
+from urllib.parse import quote
+
+import requests
+
+from ..spec import AbstractBufferedFile, AbstractFileSystem
+from ..utils import infer_storage_options
 
 logger = logging.getLogger("webhdfs")
 
 
 class WebHDFS(AbstractFileSystem):
     """
-    Interface to HDFS over HTTP
+    Interface to HDFS over HTTP using the WebHDFS API. Supports also HttpFS gateways.
 
     Three auth mechanisms are supported:
 
@@ -44,7 +49,7 @@ class WebHDFS(AbstractFileSystem):
         kerb_kwargs=None,
         data_proxy=None,
         use_https=False,
-        **kwargs
+        **kwargs,
     ):
         """
         Parameters
@@ -66,7 +71,7 @@ class WebHDFS(AbstractFileSystem):
             the user in who's name actions are taken
         kerb_kwargs: dict
             Any extra arguments for HTTPKerberosAuth, see
-            https://github.com/requests/requests-kerberos/blob/master/requests_kerberos/kerberos_.py
+            `<https://github.com/requests/requests-kerberos/blob/master/requests_kerberos/kerberos_.py>`_
         data_proxy: dict, callable or None
             If given, map data-node addresses. This can be necessary if the
             HDFS cluster is behind a proxy, running on Docker or otherwise has
@@ -120,7 +125,7 @@ class WebHDFS(AbstractFileSystem):
         args = kwargs.copy()
         args.update(self.pars)
         args["op"] = op.upper()
-        logger.debug(url, method, args)
+        logger.debug("sending %s with %s", url, method)
         out = self.session.request(
             method=method.upper(),
             url=url,
@@ -128,12 +133,22 @@ class WebHDFS(AbstractFileSystem):
             data=data,
             allow_redirects=redirect,
         )
-        if out.status_code == 404:
-            raise FileNotFoundError(path)
-        if out.status_code == 403:
-            raise PermissionError(path or "")
-        if out.status_code == 401:
-            raise PermissionError  # not specific to path
+        if out.status_code in [400, 401, 403, 404, 500]:
+            try:
+                err = out.json()
+                msg = err["RemoteException"]["message"]
+                exp = err["RemoteException"]["exception"]
+            except (ValueError, KeyError):
+                pass
+            else:
+                if exp in ["IllegalArgumentException", "UnsupportedOperationException"]:
+                    raise ValueError(msg)
+                elif exp in ["SecurityException", "AccessControlException"]:
+                    raise PermissionError(msg)
+                elif exp in ["FileNotFoundException"]:
+                    raise FileNotFoundError(msg)
+                else:
+                    raise RuntimeError(msg)
         out.raise_for_status()
         return out
 
@@ -145,7 +160,7 @@ class WebHDFS(AbstractFileSystem):
         autocommit=True,
         replication=None,
         permissions=None,
-        **kwargs
+        **kwargs,
     ):
         """
 
@@ -226,10 +241,14 @@ class WebHDFS(AbstractFileSystem):
     def ukey(self, path):
         """Checksum info of file, giving method and result"""
         out = self._call("GETFILECHECKSUM", path=path, redirect=False)
-        location = self._apply_proxy(out.headers["Location"])
-        out2 = self.session.get(location)
-        out2.raise_for_status()
-        return out2.json()["FileChecksum"]
+        if "Location" in out.headers:
+            location = self._apply_proxy(out.headers["Location"])
+            out2 = self.session.get(location)
+            out2.raise_for_status()
+            return out2.json()["FileChecksum"]
+        else:
+            out.raise_for_status()
+            return out.json()["FileChecksum"]
 
     def home_directory(self):
         """Get user's home directory"""
@@ -317,6 +336,23 @@ class WebHDFS(AbstractFileSystem):
             recursive="true" if recursive else "false",
         )
 
+    def rm_file(self, path, **kwargs):
+        self.rm(path)
+
+    def cp_file(self, lpath, rpath, **kwargs):
+        with self.open(lpath) as lstream:
+            tmp_fname = "/".join([self._parent(rpath), f".tmp.{secrets.token_hex(16)}"])
+            # Perform an atomic copy (stream to a temporory file and
+            # move it to the actual destination).
+            try:
+                with self.open(tmp_fname, "wb") as rstream:
+                    shutil.copyfileobj(lstream, rstream)
+                self.mv(tmp_fname, rpath)
+            except BaseException:  # noqa
+                with suppress(FileNotFoundError):
+                    self.rm(tmp_fname)
+                raise
+
     def _apply_proxy(self, location):
         if self.proxy and callable(self.proxy):
             location = self.proxy(location)
@@ -344,7 +380,7 @@ class WebHDFile(AbstractBufferedFile):
             self.path = "/".join([tempdir, str(uuid.uuid4())])
 
     def _upload_chunk(self, final=False):
-        """ Write one part of a multi-block file upload
+        """Write one part of a multi-block file upload
 
         Parameters
         ==========
@@ -352,35 +388,47 @@ class WebHDFile(AbstractBufferedFile):
             This is the last block, so should complete file, if
             self.autocommit is True.
         """
-        out = self.fs.session.post(self.location, data=self.buffer.getvalue())
+        out = self.fs.session.post(
+            self.location,
+            data=self.buffer.getvalue(),
+            headers={"content-type": "application/octet-stream"},
+        )
         out.raise_for_status()
         return True
 
     def _initiate_upload(self):
         """ Create remote file/upload """
+        kwargs = self.kwargs.copy()
         if "a" in self.mode:
             op, method = "APPEND", "POST"
         else:
             op, method = "CREATE", "PUT"
-            if self.fs.exists(self.path):
-                # no "truncate" or "create empty"
-                self.fs.rm(self.path)
-        out = self.fs._call(op, method, self.path, redirect=False, **self.kwargs)
+            kwargs["overwrite"] = "true"
+        out = self.fs._call(op, method, self.path, redirect=False, **kwargs)
         location = self.fs._apply_proxy(out.headers["Location"])
         if "w" in self.mode:
             # create empty file to append to
-            out2 = self.fs.session.put(location)
+            out2 = self.fs.session.put(
+                location, headers={"content-type": "application/octet-stream"}
+            )
             out2.raise_for_status()
         self.location = location.replace("CREATE", "APPEND")
 
     def _fetch_range(self, start, end):
+        start = max(start, 0)
+        end = min(self.size, end)
+        if start >= end or start >= self.size:
+            return b""
         out = self.fs._call(
             "OPEN", path=self.path, offset=start, length=end - start, redirect=False
         )
         out.raise_for_status()
-        location = out.headers["Location"]
-        out2 = self.fs.session.get(self.fs._apply_proxy(location))
-        return out2.content
+        if "Location" in out.headers:
+            location = out.headers["Location"]
+            out2 = self.fs.session.get(self.fs._apply_proxy(location))
+            return out2.content
+        else:
+            return out.content
 
     def commit(self):
         self.fs.mv(self.path, self.target)
